"use strict";
var Bluebird = require('bluebird');
var lodash_1 = require('lodash');
function buildSynchronizer(spec, options) {
    var pool = null;
    function synchronizer(transaction, item, rip, context) {
        if (item.action === 'INSERT') {
            return insert(pool, spec.name, (options && options.insertErrorHandler) || undefined, (options && options.insertInterceptor) || undefined, transaction, item, rip, context);
        }
        if (item.action === 'UPDATE') {
            return update(pool, spec.name, (options && options.updateErrorHandler) || undefined, (options && options.updateInterceptor) || undefined, transaction, item, rip, context);
        }
        if (item.action === 'REMOVE') {
            return remove(pool, spec.name, (options && options.removeErrorHandler) || undefined, (options && options.removeInterceptor) || undefined, transaction, item, rip, context);
        }
        return Bluebird.reject(new Error('provided item is invalid'));
    }
    function initialize(p) {
        pool = p;
    }
    return { synchronizer: synchronizer, initialize: initialize };
}
exports.buildSynchronizer = buildSynchronizer;
var INJECTED_MASTER_RESOLUTION_KEY = '__INJECTED_MASTER_RESOLUTION_KEY';
function insert(pool, modelName, errorHandler, interceptor, transaction, item, rip, context) {
    var model = pool.models[modelName];
    if (!model) {
        throw new Error("model " + modelName + " does not exist");
    }
    function defaultInsertBehavior() {
        var data = lodash_1.assign({}, item.data);
        var masters = [];
        model.spec.relations.master.forEach(function (master) {
            var masterData = data[master.alias];
            if (masterData) {
                var masterModel = pool.models[master.model];
                if (!masterModel.synchronizer) {
                    throw new Error("model " + masterModel.name + " must have a synchronizer");
                }
                masters.push(masterModel.synchronizer(transaction, (_a = {
                        action: 'INSERT',
                        data: masterData,
                        itemId: item.itemId,
                        model: masterModel.name
                    },
                    _a[INJECTED_MASTER_RESOLUTION_KEY] = function (id) { return data[master.key] = id; },
                    _a
                ), rip, context));
                data[master.alias] = null;
            }
            var _a;
        });
        var modelCreationPromise = Bluebird.all(masters).then(function () {
            if (!model.sequelizeModel) {
                throw new Error("model " + model.name + " must have a sequelize model");
            }
            return model.sequelizeModel.create(data, {
                transaction: transaction,
                fields: model.allowedAttributes,
            });
        });
        var promise = modelCreationPromise.then(function (instance) {
            if (!model.primaryKey) {
                throw new Error("model " + model.name + " must have a primaryKey");
            }
            if (item[INJECTED_MASTER_RESOLUTION_KEY]) {
                item[INJECTED_MASTER_RESOLUTION_KEY](instance[model.primaryKey]);
            }
            var details = [];
            var aliasModelMap = model.aliasModelMap;
            var aliasKeyMap = model.aliasKeyMap;
            Object.keys(aliasModelMap).forEach(function (alias) {
                var nested = data[alias];
                var targetModel = pool.models[aliasModelMap[alias]];
                if (!targetModel) {
                    throw new Error("model " + aliasModelMap[alias] + " does not exist");
                }
                if (nested) {
                    if (Array.isArray(nested)) {
                        nested.forEach(function (nestedItem) {
                            if (!targetModel.synchronizer) {
                                throw new Error("model " + targetModel.name + " must have a synchronizer");
                            }
                            if (!model.primaryKey) {
                                throw new Error("model " + model.name + " must have a primaryKey");
                            }
                            var nestedInsertData = lodash_1.assign({}, nestedItem, (_a = {}, _a[aliasKeyMap[alias]] = instance[model.primaryKey], _a));
                            details.push(targetModel.synchronizer(transaction, {
                                action: 'INSERT',
                                data: nestedInsertData,
                                itemId: item.itemId,
                                model: aliasModelMap[alias],
                            }, rip, context));
                            var _a;
                        });
                    }
                    else {
                        if (!targetModel.synchronizer) {
                            throw new Error("model " + targetModel.name + " must have a synchronizer");
                        }
                        if (!model.primaryKey) {
                            throw new Error("model " + model.name + " must have a primaryKey");
                        }
                        var nestedInsertData = lodash_1.assign({}, nested, (_a = {}, _a[aliasKeyMap[alias]] = instance[model.primaryKey], _a));
                        details.push(targetModel.synchronizer(transaction, {
                            action: 'INSERT',
                            data: nestedInsertData,
                            itemId: item.itemId,
                            model: aliasModelMap[alias],
                        }, rip, context));
                    }
                }
                var _a;
            });
            return details.length > 0 ? Bluebird.all(details) : Bluebird.resolve(undefined);
        }).then(lodash_1.noop);
        if (typeof errorHandler === 'function') {
            return promise.catch(function (error) { return errorHandler(item, error, rip); });
        }
        return promise;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, rip, pool, context).then(function (resume) {
            if (resume) {
                return defaultInsertBehavior();
            }
            return;
        });
    }
    return defaultInsertBehavior();
}
function update(pool, modelName, errorHandler, interceptor, transaction, item, rip, context) {
    var model = pool.models[modelName];
    if (!model) {
        throw new Error("model " + modelName + " does not exist");
    }
    function defaultUpdateBehavior() {
        if (!model.sequelizeModel) {
            throw new Error("model " + model.name + " must have a sequelize model");
        }
        if (!model.primaryKey) {
            throw new Error("model " + model.name + " must have a primaryKey");
        }
        var promise = model.sequelizeModel.update(item.data, {
            transaction: transaction,
            fields: buildUpdateAttributes(model.mutableAttributes, item.attributes),
            validate: true,
            where: (_a = {}, _a[model.primaryKey] = item.data[model.primaryKey], _a),
        }).then(lodash_1.noop);
        if (typeof errorHandler === 'function') {
            return promise.catch(function (error) { return errorHandler(item, error, rip); });
        }
        return promise;
        var _a;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, rip, pool, context).then(function (resume) {
            if (resume) {
                return defaultUpdateBehavior();
            }
            return;
        });
    }
    return defaultUpdateBehavior();
}
function remove(pool, modelName, errorHandler, interceptor, transaction, item, rip, context) {
    var model = pool.models[modelName];
    if (!model) {
        throw new Error("model " + modelName + " does not exist");
    }
    function defaultRemoveBehavior() {
        if (!model.sequelizeModel) {
            throw new Error("model " + model.name + " must have a sequelize model");
        }
        var attributes = lodash_1.intersection(Object.keys(item.identifiers), model.allowedAttributes);
        var filteredIdentifiers = {};
        attributes.forEach(function (identifier) {
            filteredIdentifiers[identifier] = item.identifiers[identifier];
        });
        var promise = model.sequelizeModel.destroy({
            transaction: transaction,
            where: filteredIdentifiers,
        }).then(lodash_1.noop);
        if (typeof errorHandler === 'function') {
            return promise.catch(function (error) { return errorHandler(item, error, rip); });
        }
        return promise;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, rip, pool, context).then(function (resume) {
            if (resume) {
                return defaultRemoveBehavior();
            }
            return;
        });
    }
    return defaultRemoveBehavior();
}
function buildUpdateAttributes(allowedAttributes, updateAttributes) {
    if (Array.isArray(allowedAttributes)) {
        return lodash_1.intersection(allowedAttributes, updateAttributes);
    }
    return updateAttributes;
}
